/**
 *  project1.cpp
 *	CS 6610 - Advanced Computer Graphics I
 *  Zach Gildersleeve
 *	Sept. 21, 2006
 *	
 *	Creates an OpenGL implementation of an animated walking
 *	robot, and uses a GLUI interface to provide controls to
 *	start and stop the robot, as well as to control the rate
 *	and direction of its movement.  Detailed documentation
 *	for this project can be found in "project1.html"
 */

#include "stdafx.cpp"

int main_window;					// the main window name

GLUI				*glui;			// pointers for all of the glui controls
GLUI_Rollout		*object_rollout;
GLUI_RadioGroup		*object_type_radio;
GLUI_Rotation		*object_rotation;
GLUI_Translation	*object_xz_trans;

GLUI_Rollout		*anim_rollout;
GLUI_Button			*action_button;

GLUI_Checkbox		*draw_floor_check;
GLUI_Checkbox		*draw_object_check;
GLUI_Checkbox		*use_depth_buffer;

float objectRotation[16];			// GLUI rotation arcball matrix
float objectXZTranslation[2];		// GLUI XZ translation matrix
int drawFloor = 1;					// state to control drawing floor				
int depthBuffer = 1;				// state to control depth buffer use
int currentButton = -1;				// current GLUI button selected
float resetRotation[] = {1.0, 0.0, 0.0, 0.0,	// identity rotation matrix
						 0.0, 1.0, 0.0, 0.0,
						 0.0, 0.0, 1.0, 0.0,
						 0.0, 0.0, 0.0, 1.0};

GLfloat armSwing = STARTING_ROTATION;			// maximum swing rotation for shoulder joint
GLfloat lowerArmSwing = -STARTING_ROTATION;		// maximum swing rotation for elbow joint
GLfloat armSpeed = 5.0;							// how quickly arms swing
	
GLfloat legSwing = STARTING_ROTATION;			// maximum swing rotation for hip joint
GLfloat lowerLegSwingMax = STARTING_ROTATION;	// maximum swing rotation for knee joint
GLfloat lowerLegSwing = STARTING_ROTATION;		// knee joint rotation multiplier
GLfloat legSpeed = 5.0;							// how quickly legs swing

GLfloat forward = 0.0;							// forward movement
GLfloat walkSpeed = 1.0;						// how fast robot walks
GLfloat rotation = 0.0;

GLint degree[NUMBER_OF_JOINTS + 1] = {0};		// array that holds rotation value of joints
												// offset to match '1' based joint naming convention

GLint currentSelection = 0;						// holds name of current body geometry selected				
GLint animation = 1;							// animation on / off toggle
GLint reset = 0;								// reset animation and robot toggle

GLfloat modelview[16];							// stores the current modelview matrix
GLfloat modelviewTemp[16];						// stores a temp modelview matrix
GLfloat tempX;									// X coordinate of robot
GLfloat tempZ;									// Y coordinate of robot

/**
 *	Sets up OpenGL scene and builds necessary lights.
 */
GLvoid Init(GLvoid)
{
	GLfloat  ambientLight[] = { 0.2f, 0.2f, 0.2f, 1.0f };		// color of overall ambient light
	GLfloat  diffuseLight[] = { 0.5f, 0.5f, 0.5f, 1.0f };		// color of light
	GLfloat	 lightPos[] = { 25.0f, 25.0f, 25.0f, 0.0};			// position of light
	GLfloat  spotDir[] = {0.0, 0.0, 0.0, 0.0};					// direction of light

	glEnable(GL_DEPTH_TEST);
	glShadeModel(GL_SMOOTH);
		
	glEnable(GL_LIGHTING);
		glEnable(GL_LIGHT0);	
		glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambientLight);
		glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuseLight);
		glLightfv(GL_LIGHT0, GL_POSITION, lightPos);
		glLightfv(GL_LIGHT0, GL_SPOT_DIRECTION, spotDir);
		
	glEnable(GL_COLOR_MATERIAL);
	glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
	
	glClearColor(0.0, 0.0, 0.0, 0.0);
}

/**
 *	Provides animation through offset redisplay
 *
 *	@param	value	frame offset time
 */
void TimerFunc(int value)
{
	if (glutGetWindow() != main_window)
		glutSetWindow(main_window);

	glutPostRedisplay();
	glutTimerFunc(100, TimerFunc, 1);
}

/**
 *	GLUT reshape function, redraws window in event of window resize.
 *	Also contains camera positional calls.
 *
 *	@param	Width	the current width of window
 *	@param	Height	the current height of window
 */
void Reshape(int Width, int Height)
{
    glViewport (0, 0, (GLsizei) Width, (GLsizei) Height);
	
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(45.0, (GLfloat) Width / (GLfloat) Height, 1.0, 200.0);
	
    glMatrixMode(GL_MODELVIEW);	
    glLoadIdentity();
	gluLookAt(2.0, 2.0, 25.0, -5.0, 0.0, 0.0, 0.0, 1.0, 0.0);
}

/**
 *	GLUI callback function, implements user interaction.
 *	Reset case resets all animation and user selected robot control.
 *
 *	@param	control		the callback name to process
 */
void glui_cb(int control)
{
	switch(control)
	{
		case CB_DEPTH_BUFFER:
		{
			if (depthBuffer)
				glEnable(GL_DEPTH_TEST);
			else
				glDisable(GL_DEPTH_TEST);
			break;
		}
		case CB_ACTION_BUTTON:
		{
			if (animation == 1)
			{
				animation = 0;
				action_button->set_name("Start");
			}
			else
			{
				reset = 0;
				animation = 1;
				action_button->set_name("Stop");
			}
			break;
		}
		case CB_RESET:
		{
			if (!animation)
				animation = 1;

			glui_cb(CB_ACTION_BUTTON);

			for (int i = 0; i < NUMBER_OF_JOINTS + 1; i++)
			{
				degree[i] = 0;
			}
			reset = 1;
			forward = 0.0;

			for (int i = 0; i < 16; i++)
			{
				objectRotation[i] = resetRotation[i];
			}

			lowerLegSwing = lowerLegSwingMax = legSwing = armSwing = STARTING_ROTATION;
			object_rotation->reset();
			objectXZTranslation[0] = 0;
			objectXZTranslation[1] = 0;
			lowerArmSwing = -STARTING_ROTATION;
		}
		break;
	}
	glutPostRedisplay();
}

/**
 *	Provides means to select robot geometry by means of GLUT mouse functionality.
 *
 *	@param	button	the user's button processed
 *	@param	state	the state of the button
 *	@param	x		the x position of the mouse
 *	@param	y		the y position of the mouse
 */
void selectObjects(int button, int state, int x, int y)
{
	GLuint selectBuf[BUFFER_SIZE];		// array of returned hit records
	GLint hits;							// number of hits returned
	GLint viewport[4];					// array to hold current viewport values
	
	if (state == GLUT_DOWN)
		currentButton = button;
	else
	{
		if (button == currentButton)
			currentButton = -1;
	}

	if (button != GLUT_LEFT_BUTTON || state != GLUT_DOWN)
	{
		currentSelection = 0;
		return;
	}

	glGetIntegerv(GL_VIEWPORT, viewport);

	glSelectBuffer(BUFFER_SIZE, selectBuf);
	(void) glRenderMode(GL_SELECT);
	glInitNames();
	glPushName(0);
	
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	gluPickMatrix((GLdouble) x, (GLdouble) (viewport[3] - y), 1.0, 1.0, viewport);
    gluPerspective(45.0, (GLfloat) (viewport[2] - viewport[0])/(GLfloat)(viewport[3] - viewport[1]), 0.1, 200.0);

	glMatrixMode(GL_MODELVIEW);
	drawRobot(GL_SELECT);
	
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	Reshape(viewport[2], viewport[3]);
	
	hits = glRenderMode(GL_RENDER);
	if (hits > 0)
	{
		processHits(hits, selectBuf);	
		animation = 0;	
	}
}

/**
 *	Processes the hits when in GL selection mode.  Stores the top-most
 *	object selected.
 *
 *	@param	hits		the number of hits returned
 *	@param	buffer[]	the selection buffer		
 */
GLvoid processHits(GLint hits, GLuint buffer[])
{
	GLuint *ptr;	// local values to process hit buffer
	GLint names;

	ptr = (GLuint *) buffer;

	if (!animation)
		animation = 1;

	glui_cb(CB_ACTION_BUTTON);

	for (int i = 0; i < hits; i++)
	{
		names = *ptr;
		ptr++;		// skip over number of names on name stack
		ptr++;		// skip over min z-value
		ptr++;		// skip over max z-value
		
		for (int j = 0; j < names; j++)
		{
			currentSelection = *ptr;
			ptr++;
		}
	}
}

/**
 *	GLUT mouse motion function.  Attempts to constrain mouse feeback
 *	in selection mode to rotate joints more naturally.
 *
 *	@param	x	the current mouse x coordinate
 *	@param	y	the current mouse y coordinate
 */
void motion(int x, int y)
{
	degree[currentSelection] = static_cast<GLint> (sqrt(static_cast<GLfloat>(x * x + y * y)));
}

/**
 *	Creates a checkered ground out of geometry.
 */
GLvoid drawGround(GLvoid)
{
	GLint x, z;
	GLint counter = 0;
	
	for (x = -GROUND_SIZE; x < GROUND_SIZE; x += CHECK_SIZE)
	{
		for (z = -GROUND_SIZE; z < GROUND_SIZE; z += CHECK_SIZE)
		{
			glColor3f(1.0, 1.0, 1.0);
			if (counter % 2 == 0)
			{
				glColor3f(0.2, 0.2, 1.0);
			}
			glBegin(GL_QUADS);
				glNormal3d(0, 1, 0);
				glVertex3f(x, FLOOR_HEIGHT, z);
				glVertex3f(x, FLOOR_HEIGHT, z + CHECK_SIZE);
				glVertex3f(x + CHECK_SIZE, FLOOR_HEIGHT, z + CHECK_SIZE);
				glVertex3f(x + CHECK_SIZE, FLOOR_HEIGHT, z);
			glEnd();			
			counter++;
		}
		counter++;
	}
}

/**
 *	Draws the robot using hierarchial modeling procedures.
 *  Robot consists of an immobile head and tatoo attached to a torso,
 *	and arms and legs with elbow and knee joints
 *
 *	@param	mode	the current OpenGL matrix mode
 */
GLvoid drawRobot(GLenum mode)
{ 
	glPushMatrix();
	
	moveRobot(walkSpeed); 	

		glPushMatrix();							// Body of Robot
			glColor3f(0.5, 0.9, 0.5);
			glScalef(1.0, 2.0, 1.0);

			glutSolidCube(1.0);
		glPopMatrix();
		
		glPushMatrix();							// Tatoo
			glScalef(0.5, 0.5, 0.5);
			glTranslatef(-0.8, 0.4, 1.1);
			glLineWidth(3.0);
			glColor3f(1.0, 0.0, 0.0);
						
			glBegin(GL_LINE_STRIP);
				glVertex2f(0.0, 0.8);
				glVertex2d(0, 1);
				glVertex2d(1, 1);
				glVertex2d(0, 0);
				glVertex2d(1, 0);
				glVertex2f(1.0, 0.2);
			glEnd();
			
			glTranslatef(0.5, -0.7, 0.0);
			
			glBegin(GL_LINE_STRIP);
				glVertex2f(1.0, 0.8);
				glVertex2d(1, 1);
				glVertex2d(0, 1);
				glVertex2d(0, 0);
				glVertex2d(1, 0);
				glVertex2f(1.0, 0.5);
				glVertex2f(0.5, 0.5);
			glEnd();
		glPopMatrix();
		
		glColor3f(0.5, 0.9, 0.5);
		
		glPushMatrix();							// Head
			glTranslatef(0.0, 1.5, 0.0);
			glutSolidCube(0.65);
		glPopMatrix();
		
		glPushMatrix();							// Upper left arm
			if(mode == GL_SELECT)
				glLoadName(UPPER_LEFT_ARM);
				
			glTranslatef(0.0, 1.0, 0.0);
			glRotatef(-(armSwing) + degree[1], 1.0, 0.0, 0.0);
			
			glPushMatrix();
				glScalef(0.35, 1.0, 0.35);
				glTranslatef(2.25, -0.5, 0.0);
				glutSolidCube(1.0);
			glPopMatrix();
			
			if(mode == GL_SELECT)				// Lower left arm
				glLoadName(LOWER_LEFT_ARM);

			glTranslatef(0.0, -1.0, 0.0);
			glRotatef((armSwing < 5)?(lowerArmSwing / 2) + degree[2]:
					  (lowerArmSwing * (armSwing / 10)) + degree[2], 1.0, 0.0, 0.0);
			glScalef(0.35, 1.0, 0.35);
			glTranslatef(2.25, -0.55, 0.0);
			glutSolidCube(1.0);
		glPopMatrix();
		
		glPushMatrix();							// Upper right arm
			if(mode == GL_SELECT)
				glLoadName(UPPER_RIGHT_ARM);
				
			glTranslatef(0.0, 1.0, 0.0);
			glRotatef(armSwing + degree[3], 1.0, 0.0, 0.0);
			
			glPushMatrix();
				glScalef(0.35, 1.0, 0.35);
				glTranslatef(-2.25, -0.5, 0.0);
				glutSolidCube(1.0);
			glPopMatrix();
			
			if(mode == GL_SELECT)				// Lower right arm
				glLoadName(LOWER_RIGHT_ARM);
				
			glTranslatef(0.0, -1.0, 0.0);

			glRotatef((armSwing < -5)?(lowerArmSwing * (armSwing / -10)) + degree[4]:
					  (lowerArmSwing / 2) + degree[4], 1.0, 0.0, 0.0);
			glScalef(0.35, 1.0, 0.35);
			glTranslatef(-2.25, -0.55, 0.0);
			glutSolidCube(1.0);
		glPopMatrix();
		
		glPushMatrix();							// Upper left leg
			if(mode == GL_SELECT)
				glLoadName(UPPER_LEFT_LEG);
				
			glTranslatef(0.75, -1.0, 0.0);
			glRotatef(legSwing + degree[5], 1.0, 0.0, 0.0);
			
			glPushMatrix();
				glScalef(0.45, 1.0, 0.45);
				glTranslatef(0.0, -.65, 0.0);
				glutSolidCube(1.0);
			glPopMatrix();
			
			if(mode == GL_SELECT)				// Lower left leg
				glLoadName(LOWER_LEFT_LEG);
			glTranslatef(0.0, -1.0, 0.0);
			glRotatef((legSwing < 5.0)?(lowerLegSwing * (fabs(legSwing) / 20.0)) + degree[6]:
					  (lowerLegSwing * (fabs(legSwing) / 10.0)) + degree[6], 1.0, 0.0, 0.0);
			glScalef(0.45, 1.0, 0.45);
			glTranslatef(0.0, -0.65, 0.0);
			glutSolidCube(1.0);
		glPopMatrix();
		
		glPushMatrix();							// Upper right leg
			if(mode == GL_SELECT)
				glLoadName(UPPER_RIGHT_LEG);
			glTranslatef(-0.75, -1.0, 0.0);
			glRotatef(-legSwing + degree[7], 1.0, 0.0, 0.0);
			
			glPushMatrix();
				glScalef(0.45, 1.0, 0.45);
				glTranslatef(0.0, -0.65, 0.0);
				glutSolidCube(1.0);
			glPopMatrix();
			
			if(mode == GL_SELECT)				// Lower right leg
				glLoadName(LOWER_RIGHT_LEG);

			glTranslatef(0.0, -1.0, 0.0);
			glRotatef((legSwing < 5.0)?(lowerLegSwing * (fabs(legSwing) / 10.0)) + degree[8]:
					  (lowerLegSwing * (fabs(legSwing) / 20.0)) + degree[8], 1.0, 0.0, 0.0);
			glScalef(0.45, 1.0, 0.45);
			glTranslatef(0.0, -0.65, 0.0);
			glutSolidCube(1.0);
		glPopMatrix();	
	glPopMatrix();
	
	if (mode == GL_RENDER)
		glutSwapBuffers();
}

/**
 *	Animates the robot by incrementing the arm and leg
 *  rotation appropriately, and incrementing the forward movement
 *	
 *	@param speedScale	the scale of how fast the robot moves
 */
GLvoid moveRobot(GLfloat speedScale)
{	
	glTranslatef(objectXZTranslation[0], 0.0, -objectXZTranslation[1]);

	if (!reset)
	{
		glGetFloatv(GL_MODELVIEW_MATRIX, modelview);

		tempX = modelview[12];
		tempZ = modelview[14];

		modelview[12] = 0.0;
		modelview[14] = 0.0;
		
		glLoadMatrixf(modelview);

		rotation = acos(objectRotation[0]) * 360.0;
		glRotatef(rotation, 0.0, 1.0, 0.0);

		glGetFloatv(GL_MODELVIEW_MATRIX, modelview);

		modelview[12] = tempX;
		modelview[14] = tempZ;

		glLoadMatrixf(modelview);
	}

	if (animation)
	{
		if (armSwing < lowerArmSwing || armSwing > -(lowerArmSwing))
		{
			armSpeed = -(armSpeed) ;		
		}
		
		armSwing += (armSpeed * speedScale);

		if (legSwing > lowerLegSwingMax || legSwing < -(lowerLegSwingMax))
		{
			legSpeed = -(legSpeed);		
		}
		if (speedScale < 1.0)
		{
			legSwing += (legSpeed * speedScale);		
		}
		else
		{
			legSwing += (legSpeed * speedScale * speedScale);
		}
		
		forward += (0.1 * speedScale * speedScale);
	}

	glTranslatef(0.0, 0.0, forward);
}

/**
 *	GLUT display function, clears the necessary buffers and draws
 *	the necessary geometry.
 */
void Display(void)
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glPushMatrix();
		if (drawFloor)
			drawGround();
	glPopMatrix();

	drawRobot(GL_RENDER);
}

/**
 *	The main function.  Creates a window and implements the
 *	GLUI interface.
 */
int main(int argc, char** argv)
{
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
    glutInitWindowSize(WINDOW_WIDTH, WINDOW_HEIGHT); 
    glutInitWindowPosition (100, 100);
    main_window = glutCreateWindow ("ROBOT!");
	
	Init();

    glutDisplayFunc(Display); 
    glutReshapeFunc(Reshape);
	glutMouseFunc(selectObjects);
	glutMotionFunc(motion);
	glutTimerFunc(250, TimerFunc, 1);

	glui = GLUI_Master.create_glui_subwindow(main_window, GLUI_SUBWINDOW_LEFT);
	objectXZTranslation[0] = 0;
	objectXZTranslation[1] = 0;

	glui->add_button("Quit", 0, (GLUI_Update_CB)exit);

	glui->add_statictext("");

	object_rollout = glui->add_rollout("Move It!");
	GLUI_Panel *transform_panel = glui->add_panel_to_panel(object_rollout, "", GLUI_PANEL_NONE);
	object_rotation = glui->add_rotation_to_panel(transform_panel, "Rotation", objectRotation);
	object_rotation->reset();
	glui->add_column_to_panel(transform_panel, false);
	object_xz_trans = glui->add_translation_to_panel(transform_panel, "Translate XZ", GLUI_TRANSLATION_XY, objectXZTranslation);
	glui->add_column_to_panel(transform_panel, false);
	object_xz_trans->scale_factor = 0.1f;
	glui->add_button_to_panel(object_rollout, "Reset Object Transform", CB_RESET, glui_cb);

	glui->add_statictext("");

	anim_rollout = glui->add_rollout("Groove It!");
	action_button = glui->add_button_to_panel(anim_rollout, "Stop", CB_ACTION_BUTTON, glui_cb);
	GLUI_Spinner *spin_s = glui->add_spinner_to_panel(anim_rollout, "Speed", GLUI_SPINNER_FLOAT, &walkSpeed);
	spin_s->set_float_limits(0.5, 2.5);

	glui->add_checkbox("Draw Floor", &drawFloor);

	glui->add_checkbox("Use Depth Buffer", &depthBuffer, CB_DEPTH_BUFFER, glui_cb);

	glui->set_main_gfx_window(main_window);
    
    glutMainLoop();
    
    return 0;
}
